#!/bin/sh
# \
exec wish "$0" ${1+"$@"}
#
# RFSXplore --
#
#	A Tcl/Tk GUI frontend to rfstool for exploring ReiserFS partitions
#	from Windows.
#
#------------------------------------------------------------------------------
#
#	Copyright (C) 2003 Eric THIEBAUT.
#
#	This file is part of RFSXplore a Tcl/Tk GUI frontend to rfstool for
#	exploring ReiserFS partitions from Windows.
#
#	RFSXplore is free software; you can redistribute it and/or modify it
#	under the terms of the GNU General Public License as published by the
#	Free Software Foundation; either version 2 of the License, or (at
#	your option) any later version.
#
#	RFSXplore is distributed in the hope that it will be useful, but
#	WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#	General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with RFSXplore (file "LICENSE.txt" in the top source
#	directory); if not, write to the Free Software Foundation, Inc., 59
#	Temple Place, Suite 330, Boston, MA 02111-1307 USA.
#
#------------------------------------------------------------------------------

# Ensure existence of ::rfsx namespace
namespace eval ::rfsx {}

# path to rfstool executable (can be changed with the "setup" button):
set RFSX_RFSTOOL ""

# look-and-feel preferences:
set RFSX_PREFS(dirfg)    DeepSkyBlue	;# text color for directories
set RFSX_PREFS(filefg)   Black		;# text color for regular files
set RFSX_PREFS(textbg)   White		;# background color for file list
set RFSX_PREFS(textfg)   DimGray	;# text color for other files
set RFSX_PREFS(activefg) MediumBlue	;# text color for line under the mouse
set RFSX_PREFS(selectfg) Black		;# text color for selection
set RFSX_PREFS(selectbg) DarkSeaGreen1	;# background color for selection

# load some packages
package require Tk
if {[catch {package require Img}]} {
  set RFSX_IMGEXTS ".gif"
} else {
  set RFSX_IMGEXTS [list ".png" ".gif"]
}

# rfsx_change_exe - change the path to the rfstool executable
#
# Returned Value: the selected name or an empty string
proc rfsx_change_exe {{parent .}} {
  global RFSX_RFSTOOL RFSX_DIR argv0
  set script [list tk_getOpenFile -defaultextension ".exe" \
		  -parent $parent \
		  -title "Choose location of \"rfstool\" executable." \
		  -filetypes {
		    {{Executables}      {.exe}        }
		    {{All Files}        *             }
		  }]
  if {[info exists RFSX_RFSTOOL] && [file isfile $RFSX_RFSTOOL]} {
    set initialdir [file dirname $RFSX_RFSTOOL]
    set initialfile [file tail $RFSX_RFSTOOL]
    lappend script -initialdir $initialdir -initialfile $initialfile
  } else {
    lappend script -initialdir $RFSX_DIR
  }
  set filename [eval $script]
  if {$filename != ""} {
    # Check selected file name setup source code so that the same
    # executable will be used the next time.
    if {! [file isfile $filename]} {
      rfsx_dialog_error -title "Bad File" "The selected file does not exists."
      return
    }
    if {! [info exists RFSX_RFSTOOL] ||
	[string compare $RFSX_RFSTOOL $filename]} {
      set file [open $argv0]
      set code ""
      set found 0
      while {! $found && [gets $file line] >= 0} {
	set found [regsub "^(\[ \t\]*set\[ \t\]+RFSX_RFSTOOL).*$" $line \
		       "\\1 \"$filename\"" line]
	append code $line "\n"
      }
      append code [read $file]
      close $file
      if {! $found} {
	rfsx_dialog_error -title "Bad File" "\
Unable to find line

        set RFSX_RFSTOOL ...

in file \"$argv0\", you must have to edit it by hand."
	exit
      }

      file rename -force -- $argv0 $argv0.bak
      set file [open $argv0 "w"]
      puts -nonewline $file $code
      close $file

      set RFSX_RFSTOOL $filename
    }
  }
  return $filename
}
# PartitionType
# 0x0C            FAT32 (LBA)
# 0xA0            save to disk
# 0x82            Linux swap
# 0x83            Linux
proc rfsx_init {} {
  global argv0 RFSX_RFSTOOL RFSX_DIR RFSX_ICONS RFSX_VERSION

  set RFSX_VERSION 1.1

  set RFSX_DIR [file dirname $argv0]

  if {! ([info exists RFSX_RFSTOOL] && [file isfile $RFSX_RFSTOOL]) &&
      [rfsx_change_exe] == ""} {
    exit
  }

  global RFSX_TYPE
  set RFSX_TYPE(file)             "-"
  set RFSX_TYPE(directory)        "d"
  set RFSX_TYPE(characterSpecial) "c"
  set RFSX_TYPE(blockSpecial)     "b"
  set RFSX_TYPE(fifo)             "p"
  set RFSX_TYPE(link)             "l"
  set RFSX_TYPE(socket)           "s"

  global RFSX_MODE
  set RFSX_MODE(0) "---"
  set RFSX_MODE(1) "--x"
  set RFSX_MODE(2) "-w-"
  set RFSX_MODE(3) "-wx"
  set RFSX_MODE(4) "r--"
  set RFSX_MODE(5) "r-x"
  set RFSX_MODE(6) "rw-"
  set RFSX_MODE(7) "rwx"

  upvar \#0 RFSX_PARTITIONS db
  catch {unset db}
  set db(linux_partitions) {}
  set db(windows_partitions) {}
  set db(reiserfs_partitions) {}
  set drive -1
  set partition -1
  while {[catch {exec $RFSX_RFSTOOL info} result]} {
    set button [rfsx_dialog .dialog -parent . -title "Error" -icon "error" \
		    -buttons [list "Setup" "Abort"] -default 0 \
		    -message "Error while running RFSTOOL:\n\n$result\n\nYou may use the \"setup\" button below to reconfigure."]
    if {$button != 0 || [rfsx_change_exe] == ""} {
      exit
    }
  }

  foreach line [split $result "\n"] {
    set line [string trim $line]
    if {[string range $line 0 2] == "---"} {
      if {[regexp -- "^-+ PARTITION (\[0-9\]+) -+\$" $line dummy p]} {
	set partition $p
	if {$drive >= 0 && $partition >= 0} {
	  set ident ${drive}.${partition}
	  set db(${ident}:size) 0.0
	  set db(${ident}:type) -1
	  set db(${ident}:rfs)  -1
	}
      } elseif {[regexp -- "^-+ Info on .*PhysicalDrive(\[0-9\]+) -+\$" \
		     $line dummy d]} {
	set drive $d
	set partition -1
      }
    } elseif {$drive >= 0 && $partition >= 0} {
      set ident ${drive}.${partition}
      set nscan [scan $line " %\[^= \] = %d %\[^\n\]" key value tail]
      if {$nscan == 3 && $key == "PartitionLength"} {
	if {[scan $tail "(%f %s" size tmp] == 2 && $tmp == "MB)"} {
	  set db(${ident}:size) $size
	} else {
	  error "invalid \"PartitionLength\" line"
	}
      } elseif {$nscan == 2 && $key == "PartitionType"} {
	set db(${ident}:type) $value
	if {$value == 0x83} {
	  lappend db(linux_partitions) $ident
	} elseif {$value == 0xA0} {
	  lappend db(windows_partitions) $ident
	}
      } elseif {[string match "Is ReiserFS partition type *" $line]} {
	if {$db(${ident}:rfs) == -1} {
	  set db(${ident}:rfs) 1
	  lappend db(reiserfs_partitions) $ident
	} else {
	  error "corrupted output from \"rfstool\""
	}
      } elseif {[string match "Doesn't seem to be a ReiserFS partition *" \
		     $line]} {
	if {$db(${ident}:rfs) == -1} {
	  set db(${ident}:rfs) 0
	} else {
	  error "corrupted output from \"rfstool\""
	}
      }
    }
  }

  # Create images
  set dir [file join $RFSX_DIR icons]
  catch {unset RFSX_ICONS}
  foreach {name file} [list \
			   "back"       "back" \
			   "copy"       "copy" \
			   "error"      "error" \
			   "newdir"     "newdir" \
			   "info"       "info" \
			   "question"   "question"\
			   "reload"     "reload" \
			   "up"         "up" \
			   "warning"    "warning"] {
    foreach ext $::RFSX_IMGEXTS {
      set filename [file join $dir $file$ext]
      if {[file exists $filename] &&
	  ! [catch {image create photo ::rfsx::img_$name -file $filename}]} {
	set RFSX_ICONS($name) ::rfsx::img_$name
	break
      }
    }
  }
}

#
# Usage: rfsx_dir VARNAME PATH
#    List contents of PATH into array variable VARNAME.
#    PATH is a directory name in the form:
#       "DRIVE:/DIRECTORY" where DRIVE is a letter like "A" or "C"
#    or "DRIVE.PART:/DIRECTORY" DRIVE and PART are the disk and
#    partition numbers. The output arrayis filled as follows:
#      $var(:lst) = list of entries
#      $var(:dev) = current device (e.g. "A:" or "0.4:")
#      $var(:dir) = current directory
#      $var($name:type) = file type for $name
#      $var($name:mode) = permissions like "rwx--x-r-x"
#      $var($name:time) = modification time
#      $var($name:size) = file size (only for regular files)
#      $var($name:link) = path of symbolic link (only for links)
#    File type is a letter: "-" for a regular file
#                           "d" for a directory
#                           "c" for a character special file
#                           "b" for a block special file
#                           "p" for a fifo
#                           "l" for a symbolic link
#                           "s" for a socket
#
proc rfsx_dir {varname path} {
  global RFSX_TYPE RFSX_MODE RFSX_RFSTOOL
  upvar \#0 RFSX_PARTITIONS db
  upvar $varname var
  catch {unset var}

  # Split PATH in two parts: DEV and DIR whith respect to DEV
  set i [string first ":" $path]
  if {$i < 1} {
    error "illegal path \"$path\""
  }
  set dev [string toupper [string range $path 0 [expr $i - 1]]]
  set dir [string range $path [expr $i + 1] end]
  if {[string index $dir 0] != "/"} {
    set dir "/$dir"
  }
  regsub -all "/{2,}" $dir "/" dir

  set var(:lst) {}
  set var(:dev) $dev
  set var(:dir) $dir

  if {[string match "\[A-Z\]" $dev]} {
    # Windows file (use builtin Tcl commands to get file information)
    set path "$dev:$dir"
    if {[file isdirectory $path]} {
      set list {}
      foreach path [glob -nocomplain "${path}/*"  "${path}/.*"] {
	set name [file tail $path]
	if {$name != "." && $name != ".."} {
	  lappend list $path
	}
      }
    } else {
      set list $path
    }
    foreach path $list {
      file stat $path x
      set name [file tail $path]
      set lst($name) {}
      set var($name:time) $x(mtime)
      set var($name:type) [set type $RFSX_TYPE($x(type))]
      set mode $x(mode)
      set u [expr {($mode >> 6) & 7}]
      set g [expr {($mode >> 3) & 7}]
      set o [expr { $mode       & 7}]
      set var($name:mode) $RFSX_MODE($u)$RFSX_MODE($g)$RFSX_MODE($o)
      if {$type == "-"} {
	# Regular file:
	set var($name:size) $x(size)
      } elseif {$type == "l"} {
	# Symbolic link:
	set var($name:link) [file readlink $path]
      }
    }
  } elseif {! [info exists db($dev:type)]} {
    error "unknown partition \"$dev\""
  } elseif {$db($dev:rfs) == 1} {
    set status [catch {exec $RFSX_RFSTOOL -p$dev $dir} input]
    if {[string match "ERROR*" $input]} {
      # error or PATH is not a directory
      error [string trim [string range $input 6 end]]
    } else {
      foreach line [split $input "\n"] {
	# Parse output of "rfstool".  Lines not matching the expected format
	# are ignored (this is to avoid errors on verbose comments for
	# instance with Windows XP partitions).
	if {[regexp -- "^(.)(.{9}) *(\[^ \]+) (.{3}) (.{3}) (\[0-9\]+) (.{8}) (\[0-9\]+) (.*)\$" $line dummy type mode size weekday month day time year tail]} {
          if {$type == "d"} {
	    # directory (omit "." and ".." and get rid of a final space)
	    if {$tail == ". " || $tail == ".. "} continue
	    set name [string range $tail 0 end-1]
	  } elseif {$type == "l"} {
	    # symbolic link
	    if {[regexp -- "^(.*) -> (.*)\$" $tail dummy name link]} {
	      set var($name:link) $link
	    } else {
	      error "bad link name in \"$tail\""
	    }
	  } else {
	    # regular file or (non-directory) special file
	    set name $tail
	    set var($name:size) $size
	  }
	  set lst($name) {}
	  set var($name:type) $type
	  set var($name:mode) $mode
	  set var($name:time) [clock scan "$month $day, $year $time"]
	}
      }
    }
  } else {
    error "unsupported partition type for \"$dev\""
  }
  if {[info exists lst]} {
    set var(:lst) [array names lst]
  } else {
    set var(:lst) {}
  }
}


#
# {QUIT}
# ---------------------------------------------------------
# {...} {NEW_DIR} {PARENT}    |                      {SWAP}
# {PATH 1}                    |  {PATH 2}
# name drwxrwxrwx mtime size  |
#                             |

proc rfsx_set_device {id dev} {
  set ID [string toupper $id]
  set ::RFSX_${ID}_DEV $dev
  set ::RFSX_${ID}_DIR "/"
  rfsx_rescan $id
}

proc rfsx_sort_by_name id {
  upvar \#0 ::RFSX_[string toupper $id]_CONTENTS db
  set dirlist {}
  set filelist {}
  foreach name $db(:lst) {
    if {$db($name:type) == "d"} {
      lappend dirlist $name
    } else {
      lappend filelist $name
    }
  }
  set db(:lst) [concat [lsort -dictionary $dirlist] [lsort -dictionary $filelist]]
}

proc rfsx_sort_by_time id {
  upvar \#0 ::RFSX_[string toupper $id]_CONTENTS db
  set list2 {}
  set list {}
  foreach name $db(:lst) {
    lappend list2 [list $db($name:time) $name]
  }
  foreach item [lsort -index 0 -integer $list2] {
    lappend list [lindex $item 1]
  }
  set db(:lst) $list
}

proc rfsx_rescan id {
  set ID [string toupper $id]
  upvar \#0 \
      ::RFSX_${ID}_WIDGET   w \
      ::RFSX_${ID}_DEV      dev \
      ::RFSX_${ID}_DIR      dir \
      ::RFSX_${ID}_CONTENTS db
  $w configure -state normal
  $w delete 1.0 end
  if {$id == "src"} {
    rfsx_clear_selection
  }
  if {[catch {rfsx_dir db $dev$dir} result]} {
    if {[info exists ::errorInfo]} {
      set errinfo $::errorInfo
    } else {
      set errinfo {}
    }
    if {[info exists ::errorCode]} {
      set errcode $::errorCode
    } else {
      set errcode NONE
    }
    catch {unset db}
    $w configure -state disabled
    error $result $errinfo $errcode
  }
  foreach name [rfsx_sort_by_name $id] {
    set type $db($name:type)
    set mode $db($name:mode)
    set prefix "$type$mode"
    set time [clock format $db($name:time) -format "%d/%m/%Y %H:%M:%S"]
    if {$type == "-"} {
      # regular file
      $w insert end [format "%s%11d %s %s\n" \
		     $prefix $db($name:size) $time $name] fileline
    } elseif {$type == "d"} {
      $w insert end [format "%s%31s %s\n" $prefix $time $name] dirline
    } elseif {$type == "l"} {
      $w insert end [format "%s%31s %s -> %s\n" \
			 $prefix $time $name $db($name:link)]
    } else {
      $w insert end [format "%s%31s %s\n" $prefix $time $name]
    }
  }
  $w configure -state disabled
}

proc rfsx_parent id {
  set ID [string toupper $id]
  upvar \#0 ::RFSX_${ID}_DIR dir
  set dir [file dir $dir]
  rfsx_rescan $id
}

proc rfsx_newdir id {
  set ID [string toupper $id]
  upvar \#0 ::RFSX_${ID}_DIR dir ::RFSX_${ID}_DEV dev
  puts $dev
  if {[string match -nocase "\[a-z\]:" $dev]} {
    set dirname [tk_getSaveFile -parent . -title "New directory" \
		     -initialdir [file join $dev $dir]]
    if {$dirname != ""} {
      if {[catch {file mkdir $dirname} msg]} {
	rfsx_dialog_error -title "New Directory Error" \
	    "Cannot create new directory \"$dirname\":\n\n$msg\n"
      } else {
	set dirname [file nativename $dirname]
	set dev [string toupper [string range $dirname 0 1]]
	set dir [string range $dirname 2 end]
	rfsx_rescan $id
     }
    }
  } else {
    rfsx_dialog_error -title "New Directory" \
	"Cannot create directory in non-Windows partition."
  }
}

proc rfsx_copy {} {
  global RFSX_RFSTOOL
  upvar \#0 \
      ::RFSX_SRC_DEV      src_dev \
      ::RFSX_SRC_DIR      src_dir \
      ::RFSX_SRC_CONTENTS src     \
      ::RFSX_DST_DEV      dst_dev \
      ::RFSX_DST_DIR      dst_dir \
      ::RFSX_DST_CONTENTS dst
  set src_is_native [string match -nocase "\[a-z\]:" $src_dev]
  set dst_is_native [string match -nocase "\[a-z\]:" $dst_dev]
  if {! $dst_is_native} {
    rfsx_dialog_error -title "Forbidden Copy" \
	"Cannot copy to a non-Windows partition."
    return
  }
  foreach name [rfsx_get_selection] {
    set src_type $src($name:type)
    if {$src_type == "-" || $src_type == "d"} {
      set src_path [file join $src_dev $src_dir $name]
      set dst_path [file join $dst_dev $dst_dir $name]
      if {$src_type == "-"} {
	# Regular file
	if {[file exists $dst_path]} {
	  if {[rfsx_dialog .dialog -icon "question" \
		   -title "Overwrite Existing File?" \
		   -buttons [list "Yes" "No"] -default "No" \
		   -parent . -message \
		   "File \"$dst_path\" already exists, overwrite with \"$src_path\"?"] != 0} {
	    continue
	  }
	}
	if {$src_is_native} {
	  set script [list file copy -force $src_path $dst_path]
	} else {
	  set script [list exec $RFSX_RFSTOOL -p$src_dev cp \
			  [file join $src_dir $name] $dst_path]
	}
      } else {
	if {[file exists $dst_path]} {
	  rfsx_dialog_error -title "Forbidden Copy" \
	      "Cannot recursively copy directory \"$name\": destination file/directory with same name already exists."
	  continue
	}
	if {$src_is_native} {
	  set script [list file copy $src_path $dst_path]
	} else {
	  set script [list exec $RFSX_RFSTOOL -p$src_dev -r cp \
			  [file join $src_dir $name] $dst_path]
	}
      }
      if {[catch $script msg]} {
	rfsx_dialog_error -title "Copy Error" $msg
      }
    }
  }
  rfsx_clear_selection
  rfsx_rescan dst
}

proc rfsx_change_dir {id w pos} {
  set ID [string toupper $id]
  upvar \#0 \
      ::RFSX_${ID}_DEV      dev \
      ::RFSX_${ID}_DIR      dir \
      ::RFSX_${ID}_CONTENTS db
  set pos [$w index $pos]
  set index [expr int($pos) - 1]
  set name [lindex $db(:lst) $index]
  set dir [file join $dir $name]
  rfsx_rescan $id
}

# rfsx_motion - handle motion events in dirbox
proc rfsx_motion {w id xy} {
  if {[lsearch -exact [$w tag names $xy] "active"] < 0} {
    $w tag remove "active" 1.0 end
    set ID [string toupper $id]
    upvar \#0 ::RFSX_${ID}_CONTENTS db
    set pos [$w index $xy]
    set line [expr int($pos)]
    set name [lindex $db(:lst) [expr $line - 1]]
    if {$name != "" && ($db($name:type) == "d" ||
			($id == "src" && $db($name:type) == "-"))} {
      #rfsx_debug "$id:$line,$name"
      #after 1000 rfsx_debug
      $w tag add "active" $line.0 $line.end
    }
  }
}

# rfsx_click - handle mouse button events in dirbox
proc rfsx_click {w id xy action} {
  set ID [string toupper $id]
  upvar \#0 ::RFSX_${ID}_DIR dir ::RFSX_${ID}_CONTENTS db
  set pos [$w index $xy]
  set line [expr int($pos)]
  set name [lindex $db(:lst) [expr $line - 1]]
  if {$name != ""} {
    set type $db($name:type)
    if {$action == "enter"} {
      # enter directory
      if {$type == "d"} {
	set dir [file join $dir $name]
	rfsx_rescan $id
      }
    } elseif {$id == "src"} {
      if {$action == "select"} {
	# (un-)select single line
	set i [lsearch -exact [$w tag names $xy] "select"]
	$w tag remove "select" 1.0 end
	if {$i < 0 && ($type == "-" || $type == "d")} {
	  $w tag add "select" $line.0 $line.0+1line
	}
      } elseif {$action == "toggle"} {
	# toggle state of line
	if {[lsearch -exact [$w tag names $xy] "select"] >= 0} {
	  $w tag remove "select" $line.0 $line.0+1line
	} elseif {$type == "-" || $type == "d"} {
	  $w tag add "select" $line.0 $line.0+1line
	}
      } elseif {$action == "extend"} {
	# single click with shift key
	set nsel 0
	set isel -1
	set jsel -1
	foreach {i j} [$w tag ranges "select"] {
	  if {$line < [set i [expr int($i)]]} {
	    set n [expr $i - $line]
	    if {$isel < 0 || $n < $nsel} {
	      set nsel $n
	      set isel $line
	      set jsel $i
	    }
	  } elseif {$line > [set j [expr int($j) - 1]]} {
	    set n [expr $line - $j]
	    if {$isel < 0 || $n < $nsel} {
	      set nsel $n
	      set isel $j
	      set jsel $line
	    }
	  } else {
	    # nothing to do
	    set nsel 0
	    break
	  }
	}
	if {$nsel > 0} {
	  $w tag add "select" $isel.0 $jsel.0+1line
	}
      }
    }
  }
}

proc rfsx_clear_selection {} {
  $::RFSX_SRC_WIDGET tag remove "select" 1.0 end
}

proc rfsx_get_selection {} {
  set lst $::RFSX_SRC_CONTENTS(:lst)
  set sel {}
  foreach {i j} [$::RFSX_SRC_WIDGET tag ranges "select"] {
    foreach name [lrange $lst [expr int($i) - 1] [expr int($j) - 2]] {
      if {$name != ""} {
	lappend sel $name
      }
    }
  }
  #rfsx_debug [lsort -dictionary -unique $sel]
  #return
  return [lsort -dictionary -unique $sel]
}

proc rfsx_select {cmd {pos {}}} {
  set w $::RFSX_SRC_WIDGET

  upvar \#0 ::RFSX_SRC_CONTENTS db ::RFSX_SRC_LAST last
  switch -exact $cmd {
    "get" {
      set lst $db(:lst)
      set sel {}
      foreach {i j} [$w tag ranges "select"] {
	set index [expr int($i) - 1]
	set name [lindex $lst $index]
	lappend sel $name
      }
      return [lsort -dictionary -unique $sel]
    }
    "clear" {
      $w tag remove "select" 1.0 end
      return
    }
  }
  set pos [$w index $pos]
  set line [expr int($pos)]
  set name [lindex $db(:lst) [expr $line - 1]]
  if {$db($name:type) != "-"} {
    return
  }
  switch -exact $cmd {
    "new" {
      $w tag remove "select" 1.0 end
      $w tag add "select" $line.0 $line.end
      set last $line
    }
    "add" {
      $w tag add "select" $line.0 $line.end
      set last $line
    }
    "extend" {
    }
  }
}

# print a small debug message in the toolbar
proc rfsx_debug {{text ""}} {
  .bar.debug config -text $text
}

proc rfsx_text_or_icon {w imgsw img textsw text} {
  global RFSX_ICONS
  if {[info exists RFSX_ICONS($img)]} {
    $w configure $imgsw $RFSX_ICONS($img)
  } else {
    $w configure $textsw $text
  }
}

#
# double-click on directory -> change to that directory
# single click on any entry -> clear selection, select entry for copy
# <Shift-Click>   -> extend selection
# <Control-Click> -> add entry to selection
# <Control-A>     -> select all entries
#

# text tags (-exportselection false)
#   dirname  -> to bind to double click events, to colorize
#   dirline  -> the whole line with a directory
#   select   -> the whole line for selected files
#   filename -> eligible for selection
#
proc rfsx_make_dirbox {top id} {
  upvar \#0 ::RFSX_PREFS prefs
  set ID [string toupper $id]
  set frame $top.$id
  if {[winfo exists $frame]} {
    destroy $frame
  }
  if {$id == "src"} {
    set title Source
  } else {
    set title Destination
  }
  labelframe $frame -relief flat -bd 2 -relief ridge \
      -text $title -font [list Helvetica 10 bold] \
      -padx 3 -pady 3

  # Initilalize paths
  if {! [info exists ::RFSX_${ID}_DIR]} {
    set ::RFSX_${ID}_DIR "/"
  }
  if {! [info exists ::RFSX_${ID}_DEV]} {
    set ::RFSX_${ID}_DEV "C:"
  }

  # Bar with path and buttons
  set bar [frame $frame.bar -relief flat]
  set path [label $bar.path -anchor w -relief sunken -background white \
		-textvariable ::RFSX_${ID}_DIR]
  set device [menubutton $bar.device -textvariable ::RFSX_${ID}_DEV]
  set device_menu [menu $device.menu]
  $device configure -menu $device_menu

  foreach {name cmd text img} {
    rescan ::rfsx_rescan "reload" reload
    parent ::rfsx_parent "parent" up
    newdir ::rfsx_newdir "newdir" newdir
  } {
    set $name [button $bar.$name -command [list $cmd $id]]
    rfsx_text_or_icon [set $name] -image $img -text $text
  }

  foreach b {device rescan parent newdir} {
    $bar.$b configure -borderwidth 1 -relief flat
    bind $bar.$b <Enter> {%W configure -relief raised}
    bind $bar.$b <Leave> {%W configure -relief flat}
  }

  pack $device -side left  -fill y    -expand 0 -padx 1 -pady 2
  pack $newdir -side right -fill y    -expand 0 -padx 1 -pady 2
  pack $parent -side right -fill y    -expand 0 -padx 1 -pady 2
  pack $rescan -side right -fill y    -expand 0 -padx 1 -pady 2
  pack $path   -side left  -fill both -expand 1 -padx 1 -pady 2

  # text widget to display file list
  set tab 170
  set size 10
  set font [list courier $size]
  set bold [concat $font bold]
  set w [text $frame.text -relief sunken \
	     -background $prefs(textbg) -foreground $prefs(textfg) \
	     -font $font -width 60 \
	     -wrap none -undo no -cursor arrow \
	     -takefocus 1 -exportselection false -state disabled]
  $w tag configure filename -foreground $prefs(filefg) -font $bold
  $w tag configure fileline -foreground $prefs(filefg)
  $w tag configure dirname  -foreground $prefs(dirfg) -font $bold
  $w tag configure dirline  -foreground $prefs(dirfg)
  $w tag configure select   -foreground $prefs(selectfg) -background $prefs(selectbg)
  $w tag configure active   -foreground $prefs(activefg)
  $w tag raise select
  $w tag raise active

  set ::RFSX_${ID}_WIDGET $w
  bindtags $w [list DirBox $w]

  set click [list rfsx_click %W $id @%x,%y]
  if {$id == "src"} {
    bind $w <Button-1>         [concat $click "select"]
    bind $w <Control-Button-1> [concat $click "toggle"]
    bind $w <Shift-Button-1>   [concat $click "extend"]
  }

  bind $w     <Motion> [list rfsx_motion %W $id @%x,%y]
  bind $w     <Double-Button-1> [concat $click "enter"]
  bind DirBox <Leave>  [list %W tag remove "active" 1.0 end]
  bind DirBox <Button> "+focus %W"
  bind DirBox <Left>   [list %W xview scroll -1 unit]
  bind DirBox <Right>  [list %W xview scroll  1 unit]
  bind DirBox <Up>     [list %W yview scroll -1 unit]
  bind DirBox <Down>   [list %W yview scroll  1 unit]
  bind DirBox <Next>   [list %W yview scroll  1 page]
  bind DirBox <Prior>  [list %W yview scroll -1 page]
  bind DirBox <MouseWheel> {
    %W yview scroll [expr -round(%D/30.0)] units
  }

  # Add scroll bars
  set hsb [scrollbar $frame.hsb -orient horizontal -command [list $w xview]]
  set vsb [scrollbar $frame.vsb -orient vertical   -command [list $w yview]]
  $w configure -xscrollcommand [list $hsb set] -yscrollcommand [list $vsb set]

  # Create the menu for list of devices
  set devices {}
  if {$id == "src"} {
    foreach dev $::RFSX_PARTITIONS(reiserfs_partitions) {
      lappend devices ${dev}:
    }
  }
  foreach dev [string toupper [file volume]] {
    lappend devices [string index $dev 0]:
  }
  foreach dev $devices {
    $device_menu add command -label $dev \
	-command [list ::rfsx_set_device $id $dev]
  }

  if {$id == "src"} {
    set scol 0
    set wcol 1
  } else {
    set scol 1
    set wcol 0
  }
  grid $bar -column 0 -columnspan 2 -row 0 -sticky "ew"
  grid $vsb -column $scol -row 1 -sticky "ns"
  grid $hsb -column $wcol -row 2 -sticky "ew"
  grid $w   -column $wcol -row 1 -sticky "nsew"

  grid columnconfigure $frame $scol -weight 0
  grid columnconfigure $frame $wcol -weight 1
  grid rowconfigure $frame 0 -weight 0
  grid rowconfigure $frame 1 -weight 1
  grid rowconfigure $frame 2 -weight 0

  return $frame
}

proc rfsx_main {} {
  wm withdraw .
  wm title . "RFSXplore - browser for ReiserFS partitions"
  wm iconname . "RFSXplore"
  if {! [info exists ::RFSX_PARTITIONS(reiserfs_partitions)]} {
    rfsx_init
  }
  set top {}
  set bar [frame $top.bar -relief flat -bd 1 -padx 1 -pady 1]
  set about [button $bar.about -text "About" -command rfsx_about]
  set debug [label $bar.debug -text "" -relief flat -foreground firebrick -anchor center]
  set setup [button $bar.setup -text "Setup" -command rfsx_change_exe]

  set src [rfsx_make_dirbox $top src]
  set dst [rfsx_make_dirbox $top dst]
  set cpy [button $top.copy -command rfsx_copy]
  rfsx_text_or_icon $cpy -image copy -text "copy"

  foreach w [list $cpy $about $setup] {
    $w configure -borderwidth 1 -relief flat
    bind $w <Enter> {%W configure -relief raised}
    bind $w <Leave> {%W configure -relief flat}
  }

  pack $setup -side left -fill y -expand false
  pack $debug -side left -fill both -expand 1
  pack $about -side right -fill y -expand false
  grid $bar -column 0 -columnspan 3 -row 0 -sticky "ew"
  grid $src -column 0 -row 1 -sticky "nsew"
  grid $cpy -column 1 -row 1 -sticky "" -padx 3
  grid $dst -column 2 -row 1 -sticky "nsew"
  grid columnconfigure . 0 -weight 1
  grid columnconfigure . 1 -weight 0
  grid columnconfigure . 2 -weight 1
  grid rowconfigure    . 0 -weight 0
  grid rowconfigure    . 1 -weight 1

  foreach id {src dst} {
    rfsx_rescan $id
  }
  wm deiconify .
}

#------------------------------------------------------------------------------
#
# Syntax: rfsx_dialog w [options ...]
#
# Arguments:
#   w                      - Name of top-level window for dialog.
#
# Options:
#   -parent WIN            - Parent of dialog window (default is its parent).
#   -title TEXT            - Title for dialog window.
#   -script CODE           - Script to evaluate, it must be a command with a
#                            single argument which is the frame window into
#                            which some widgets can be inserted.
#   -message TEXT          - Text of message.
#   -icon NAME             - Image name, one of: "error", "info", "question"
#                            or "warning".
#   -buttons {LABEL1 ...}  - List of buttons, for instance: {ok cancel}.
#   -default INTEGER       - Index of default button (can alkso be the label of
#                            the default button).
#
# Returned Value:
#   The index (>=0) of the selected button, -1 if destroyed.
#
proc ::rfsx_dialog {w args} {
  global tcl_platform RFSX_ICONS
  variable ::tk::Priv

  # Parse options.
  array set opt {
    -script  {}
    -parent  {}
    -title   {}
    -icon    {}
    -message {}
    -buttons Ok
    -default {}
  }
  foreach {option value} $args {
    if {! [info exists opt($option)]} {
      return -code error "unknown option \"$option\""
    }
    set opt($option) $value
  }
  foreach option [array names opt] {
    set [string range $option 1 end] $opt($option)
  }

  # Check that $default was properly given
  if {$default == ""} {
    if {[llength $buttons] == 1} {
      set default 0
    } else {
      set default -1
    }
  } elseif {[string is int $default]} {
    if {$default >= [llength $buttons]} {
      return -code error "default button index greater than number of\
		    buttons specified for rfsx_dialog"
    }
  } else {
    set default [lsearch -exact $buttons $default]
  }

  # 1. Create the top-level window and divide it into top
  # and bottom parts.

  catch {destroy $w}
  toplevel $w -class Dialog
  wm title $w $title
  wm iconname $w Dialog
  wm protocol $w WM_DELETE_WINDOW { }

  # Dialog boxes should be transient with respect to their parent, so
  # that they will always stay on top of their parent window.
  # However, some window managers will create the window as withdrawn
  # if the parent window is withdrawn or iconified.  Combined with the
  # grab we put on the window, this can hang the entire application.
  # Therefore we only make the dialog transient if the parent is
  # viewable.
  if {$parent == ""} {
    set parent [winfo parent $w]
  }
  if {[winfo viewable [winfo toplevel $parent]] } {
    wm transient $w [winfo toplevel $parent]
  }
  set winsys [tk windowingsystem]
  if {$tcl_platform(platform) == "macintosh" || $winsys == "aqua"} {
    set winsys "mac"
    ::tk::unsupported::MacWindowStyle style $w dBoxProc
  }
  frame $w.bot
  frame $w.top
  if {$winsys == "x11"} {
    $w.bot configure -relief raised -bd 1
    $w.top configure -relief raised -bd 1
  }
  pack $w.bot -side bottom -fill both
  pack $w.top -side top -fill both -expand 1

  # 2. Fill the top part with image and message (use the option
  # database for -wraplength and -font so that they can be overridden
  # by the caller).

  set row -1
  if {$message != ""} {
    option add *Dialog.msg.wrapLength 3i widgetDefault
    if {$winsys == "mac"} {
      option add *Dialog.msg.font system widgetDefault
    } else {
      option add *Dialog.msg.font {Helvetica 12} widgetDefault
    }

    label $w.msg -justify left -text $message
    grid $w.msg -in $w.top -column 1 -row [incr row] -sticky "nsew" \
	-padx 3m -pady 3m

    if {$icon != "" && [info exists ::RFSX_ICONS($icon)]} {
      label $w.icon -image $::RFSX_ICONS($icon)
      grid $w.icon -in $w.top -column 0 -row $row -sticky "e" \
	  -padx 3m -pady 3m
    }
  }
  if {$script != ""} {
    frame $w.top.frame -relief sunken -bd 1
    uplevel \#0 [list $script] $w.top.frame
    grid $w.top.frame -in $w.top -column 0 -columnspan 2 -row [incr row] \
	-sticky "nsew" -padx 3m -pady 3m
  }
  if {$row >= 0} {
    grid columnconfigure $w.top 0 -weight 0
    grid columnconfigure $w.top 1 -weight 1
    if {$row == 1} {
      grid rowconfigure $w.top 0 -weight 0
      grid rowconfigure $w.top 1 -weight 1
    } else {
      grid rowconfigure $w.top 0 -weight 1
    }
  }

  # 3. Create a row of buttons at the bottom of the dialog.

  set i 0
  foreach but $buttons {
    button $w.button$i -text $but -command [list set ::tk::Priv(button) $i]
    if {$i == $default} {
      $w.button$i configure -default active
    } else {
      $w.button$i configure -default normal
    }
    grid $w.button$i -in $w.bot -column $i -row 0 -sticky ew \
	-padx 10 -pady 4
    grid columnconfigure $w.bot $i -minsize 100
    # We boost the size of some Mac buttons for l&f
    if {$winsys == "mac"} {
      set tmp [string tolower $but]
      if {$tmp == "ok" || $tmp == "cancel"} {
	grid columnconfigure $w.bot $i -minsize [expr {59 + 20}]
      }
    }
    incr i
  }

  # 4. Create a binding for <Return> on the dialog if there is a
  # default button.

  if {$default >= 0} {
    bind $w <Return> "
	[list $w.button$default] configure -state active -relief sunken
	update idletasks
	after 100
	set ::tk::Priv(button) $default
	"
  }

  # 5. Create a <Destroy> binding for the window that sets the
  # button variable to -1;  this is needed in case something happens
  # that destroys the window, such as its parent window being destroyed.

  bind $w <Destroy> {set ::tk::Priv(button) -1}

  # 6. Withdraw the window, then update all the geometry information
  # so we know how big it wants to be, then center the window in the
  # display and de-iconify it.

  wm withdraw $w
  update idletasks
  set x [expr {[winfo screenwidth $w]/2 - [winfo reqwidth $w]/2 \
		   - [winfo vrootx $parent]}]
  set y [expr {[winfo screenheight $w]/2 - [winfo reqheight $w]/2 \
		   - [winfo vrooty $parent]}]
  wm geom $w +$x+$y
  wm deiconify $w

  # 7. Set a grab and claim the focus too.

  set oldFocus [focus]
  set oldGrab [grab current $w]
  if {[string compare $oldGrab ""]} {
    set grabStatus [grab status $oldGrab]
  }
  grab $w
  if {$default >= 0} {
    focus $w.button$default
  } else {
    focus $w
  }

  # 8. Wait for the user to respond, then restore the focus and
  # return the index of the selected button.  Restore the focus
  # before deleting the window, since otherwise the window manager
  # may take the focus away so we can't redirect it.  Finally,
  # restore any grab that was in effect.

  vwait ::tk::Priv(button)
  catch {focus $oldFocus}
  catch {
    # It's possible that the window has already been destroyed,
    # hence this "catch".  Delete the Destroy handler so that
    # Priv(button) doesn't get reset by it.

    bind $w <Destroy> {}
    destroy $w
  }
  if {[string compare $oldGrab ""]} {
    if {[string compare $grabStatus "global"]} {
      grab $oldGrab
    } else {
      grab -global $oldGrab
    }
  }
  return $Priv(button)
}

proc rfsx_dialog_error args {
  if {[lindex $args 0] == "-title"} {
    set title [lindex $args 1]
    set message [join [lrange $args 2 end]]
  } else {
    set title "Error"
    set message [join $args]
  }
  rfsx_dialog .dialog -title $title -message $message -icon "error"
}

proc rfsx_dialog_info args {
  if {[lindex $args 0] == "-title"} {
    set title [lindex $args 1]
    set message [join [lrange $args 2 end]]
  } else {
    set title "Info"
    set message [join $args]
  }
  rfsx_dialog .dialog -title $title -message $message -icon "info"
}

proc rfsx_dialog_warning args {
  if {[lindex $args 0] == "-title"} {
    set title [lindex $args 1]
    set message [join [lrange $args 2 end]]
  } else {
    set title "Warning"
    set message [join $args]
  }
  rfsx_dialog .dialog -title $title -message $message -icon "warning"
}

#------------------------------------------------------------------------------
proc rfsx_about {} {
  set w .about
  catch {destroy $w}
  toplevel $w -padx 3 -pady 3
  wm withdraw $w
  wm title $w "About RFSXPlore"
  wm iconname $w "About RFSXplore"

  set size 12
  set font Helvetica
  set text [text $w.text -background [$w cget -background] -foreground Black \
		-bd 0 -relief flat -width 70 -height 12 \
		-font [list $font $size] \
		-wrap word -undo no -cursor arrow \
		-takefocus 0 -exportselection true -state normal]
  $text tag configure text -justify center -font [list $font $size]
  $text tag configure bold -justify center -font [list $font $size bold]
  $text tag configure url -justify center -font [list Courier $size]
  $text insert end \
      "RFSXplore" bold \
      " - version $::RFSX_VERSION" text \
      "\n\nRFSXplore is a graphical interface to rfstool for exploring ReiserFS partitions and copying files from ReiserFS filesystem to Windows filesystem." text \
      "\n\nRFSXplore is copyrighted  2003 by Eric Thibaut under the GNU Public License\n" text \
      "http://www-obs.univ-lyon1.fr/eric.thiebaut/soft/rfsxplore" url \
      "\n\nrfstool is copyrighted  2003, 2004 by Gerson Kurz under the GNU Public License\n" text \
      "http://q-nand-q.com" url
  $text configure -state disabled

  set ok [button $w.ok -text ok -command [list destroy $w] -width 10]
  grid $text -column 0 -row 0 -sticky "ewns"
  grid $ok -column 0 -row 1 -sticky "ns"
  grid columnconfigure $w 0 -weight 1
  grid rowconfigure $w 0 -weight 1
  grid rowconfigure $w 1 -weight 0
  wm deiconify $w
}

rfsx_main
tkwait window .
